# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import datetime
import importlib
import json
from enum import Enum
from typing import Dict

import boto3
import requests
from cloudevents.http import CloudEvent, to_structured

from .. import logger
from ..custom_exceptions import TriggerException
from . import trigger


class ProtocolType(Enum):
    """
    Enum for the various protocols accepted by the post triggers
    """

    cloudevents_http = ("post_trigger", "PostCloudEventsHttp")
    cloudevents_aws = ("post_trigger", "PostCloudEventsAws")

    def get_class(self):
        module = importlib.import_module("pyaviso.triggers." + self.value[0])
        return getattr(module, self.value[1])


class PostTrigger(trigger.Trigger):
    """
    This class implements a trigger in charge of posting the notification accordingly to the protocol selected.
    This class expects the param protocol and protocol type. The remaining fields are optional.
    """

    def __init__(self, notification: Dict, params: Dict):
        trigger.Trigger.__init__(self, notification, params)
        assert params.get("protocol") is not None, "protocol is a mandatory field"
        protocol_params = params.get("protocol")
        assert protocol_params.get("type") is not None, "protocol type is a mandatory field"
        self.protocol = ProtocolType[protocol_params.get("type").lower()].get_class()(notification, protocol_params)

    def execute(self):
        logger.info("Starting Post Trigger...'")

        # execute the specific protocol
        self.protocol.execute()

        logger.debug("Post Trigger completed")


class PostCloudEventsHttp:
    """
    This class implements a trigger in charge of translating the notification in a CloudEvents message and
    POST it to the HTTP API specified by the user.
    This class expects the params to contain the URL where to send the message to. The remaining fields are optional.
    """

    TIMEOUT_DEFAULT = 60
    TYPE_DEFAULT = "aviso"
    SOURCE_DEFAULT = "https://aviso.ecmwf.int"

    def __init__(self, notification: Dict, params: Dict):
        self.notification = notification
        assert params.get("url") is not None, "url is a mandatory field"
        self.url = params.get("url")
        self.timeout = params.get("timeout", self.TIMEOUT_DEFAULT)
        self.headers = params.get("headers", {})

        # cloudEvents specific fields
        if params.get("cloudevents"):
            self.type = params.get("cloudevents").get("type", self.TYPE_DEFAULT)
            self.source = params.get("cloudevents").get("source", self.SOURCE_DEFAULT)
        else:
            self.type = self.TYPE_DEFAULT
            self.source = self.SOURCE_DEFAULT

    def execute(self):
        # prepare the CloudEvents message

        attributes = {
            "type": self.type,
            "source": self.source,
            "time": datetime.datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ"),
        }
        data = self.notification
        event = CloudEvent(attributes, data)

        # Creates the HTTP request representation of the CloudEvents in structured content mode
        headers, body = to_structured(event)
        self.headers.update(headers)

        logger.debug(f"Sending CloudEvents notification {data}")

        # send the message
        try:
            resp = requests.post(self.url, data=body, headers=self.headers, verify=False, timeout=self.timeout)
        except Exception as e:
            logger.error("Not able to POST CloudEvents notification")
            raise TriggerException(e)
        if resp.status_code != 200:
            raise TriggerException(
                f"Not able to POST CloudEvents notification to {self.url}, "
                f"status {resp.status_code}, {resp.reason}, {resp.content.decode()}"
            )

        logger.debug("CloudEvents notification sent successfully")


class PostCloudEventsAws:
    """
    This class implements a trigger in charge of translating the notification in a CloudEvents message and send it to a
    AWS topic specified by the user.
    """

    TIMEOUT_DEFAULT = 60
    TYPE_DEFAULT = "aviso"
    SOURCE_DEFAULT = "https://aviso.ecmwf.int"

    def __init__(self, notification: Dict, params: Dict):
        self.notification = notification
        assert params.get("arn") is not None, "arn is a mandatory field"
        self.arn = params.get("arn")
        assert params.get("region_name") is not None, "region_name is a mandatory field"
        self.region_name = params.get("region_name")
        self.MessageAttributes = params.get("MessageAttributes")
        # authentication is optional or is automatically set from ~/.aws/credentials
        self.aws_access_key_id = params.get("aws_access_key_id")
        self.aws_secret_access_key = params.get("aws_secret_access_key")
        # only for FIFO topics
        self.MessageGroupId = params.get("MessageGroupId")

        # cloudEvents specific fields
        if params.get("cloudevents"):
            self.type = params.get("cloudevents").get("type", self.TYPE_DEFAULT)
            self.source = params.get("cloudevents").get("source", self.SOURCE_DEFAULT)
        else:
            self.type = self.TYPE_DEFAULT
            self.source = self.SOURCE_DEFAULT

    def execute(self):
        # prepare the AWS topic message

        attributes = {
            "type": self.type,
            "source": self.source,
            "time": datetime.datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ"),
        }
        data = self.notification
        event = CloudEvent(attributes, data)

        # Creates the HTTP request representation of the CloudEvents in structured content mode
        headers, body = to_structured(event)
        event_body = body.decode()
        event_dict = json.loads(event_body)

        # Create message for AWS topic
        aws_publish_params = {
            "TopicArn": self.arn,
            "Message": json.dumps({"default": event_body}),
            "MessageStructure": "json",
            "Subject": self.type,
        }
        if self.MessageAttributes:
            aws_publish_params["MessageAttributes"] = self.MessageAttributes
        if self.MessageGroupId:
            aws_publish_params["MessageDeduplicationId"] = event_dict["id"]
            aws_publish_params["MessageGroupId"] = self.MessageGroupId

        logger.debug(f"Sending AWS topic notification {aws_publish_params}")

        # send the message
        try:
            sns = boto3.client(
                "sns",
                region_name=self.region_name,
                aws_access_key_id=self.aws_access_key_id,
                aws_secret_access_key=self.aws_secret_access_key,
            )

            # this is the SNS standard to support
            sns.publish(**aws_publish_params)

        except Exception as e:
            logger.error("Not able to send AWS topic notification")
            raise TriggerException(e)

        logger.debug("AWS topic notification sent successfully")



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from typing import List

from .enum_handler import EnumHandler


class TimeHandler(EnumHandler):
    def __init__(self, key, values: List[str], canonic, required=False):
        super(TimeHandler, self).__init__(key, values, required)
        self._canonic = canonic

    @property
    def canonic(self) -> str:
        return self._canonic

    def canonise(self, value: any) -> str:
        return self.canonic.format(value)



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import base64


class PlainAuth:
    """
    PlainAuth implements Basic authentication.
    """

    def __init__(self, config):
        self.config = config

    def header(self):
        credentials = f"{self.config.username}:{self.config.password}"
        encoded = base64.b64encode(credentials.encode("utf-8")).decode("utf-8")
        return {"Authorization": f"Basic {encoded}", "X-Auth-Type": "plain"}



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from datetime import datetime
from typing import Dict, List, Optional

from .. import logger
from ..engine.engine_factory import EngineFactory
from ..triggers import trigger_factory as tf
from . import event_listener as el


class EventListenerFactory:
    """
    Factory class of EventListener objects. It creates them by parsing a key-value dictionary
    """

    def __init__(self, engine_factory: EngineFactory, listener_schema: Dict[str, any]):
        """
        :param engine_factory:
        :param listener_schema:
        """
        self._engine_factory = engine_factory
        self._listener_schema = listener_schema

    def create_listeners(
        self,
        listeners_dict: Dict[str, any],
        from_date: datetime = None,
        to_date: datetime = None,
        payload_key: str = None,
    ) -> List[el.EventListener]:
        """
        This method is used to parse a key-value dictionary and create a list of event listeners.

        :param listeners_dict: key-value dictionary to parse
        :param from_date: date from when to request notifications, if None it will be from now
        :param to_date: date until when to request notifications, if None it will be until now
        :param payload_key: key to use for the payload in the notification dictionary
        :return: a list of EventListener objects
        """
        listeners: List[el.EventListener] = []

        # parse the listeners dictionary
        logger.debug("Parsing the listeners file")
        assert listeners_dict is not None, "Event listeners definition cannot be empty"
        listener_list: Optional[List[Dict[str, any]]] = listeners_dict.get("listeners")
        assert listener_list is not None, "Event listeners definition must start with the keyword 'listeners'"

        # Create the engine to connect to the notification server
        engine = self._engine_factory.create_engine()

        for listen in listener_list:
            # each listener is a dictionary
            assert isinstance(listen, dict), "Wrong file structure"

            # extract the relevant listener schema
            assert "event" in listen, "Wrong file structure, 'event' could not be located"
            event_type = listen.get("event")
            assert event_type in self._listener_schema, f"Wrong schema structure, {event_type} could not be located"
            schema = self._listener_schema.get(event_type)

            # Parse the request and build the key to the forecast dataset
            assert "request" in listen, "Wrong file structure, 'request' could not be located"
            request: Optional[Dict[str, any]] = listen.get("request")

            # Parse the triggers
            triggers: Optional[List[Dict[str, any]]] = self._parse_triggers(listen)

            # create the listener
            listener = el.EventListener(event_type, engine, request, triggers, schema, from_date, to_date, payload_key)
            listeners.append(listener)

        return listeners

    def _parse_triggers(self, listener: Dict[str, any]) -> Optional[List[Dict[str, any]]]:
        """
        This method parses the triggers block and return the list of trigger types to use
        :param listener:
        :return:
        """
        # Parse the triggers
        assert "triggers" in listener, "'triggers' is a mandatory field"
        triggers: Optional[List[Dict[str, any]]] = listener.get("triggers")
        assert triggers is not None, "At least one trigger must be defined"
        # validate the types of the triggers
        for t in triggers:
            assert "type" in t, "'type' is a mandatory field in trigger"
            try:
                # early validation of the trigger type
                tf.TriggerType[t.get("type").lower()]
            except KeyError as e:
                raise KeyError(f"Trigger type {e.args[0]} not recognised")

        return triggers



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


class OpenidAuth:
    """
    OpenidAuth implements an OpenID authentication flow.

    It returns a Bearer header (using the shared secret from config.password) and adds
    an extra header "X-Auth-Type" with the value "openid".
    """

    def __init__(self, config):
        self.config = config

    def header(self):
        return {"Authorization": f"Bearer {self.config.password}", "X-Auth-Type": "openid"}



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import _thread
import os
import threading
import time
from datetime import datetime
from queue import Queue
from shutil import rmtree
from typing import Dict, List

from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer

from .. import logger
from ..authentication.auth import Auth
from ..user_config import EngineConfig
from .engine import Engine


class FileBasedEngine(Engine):
    """
    This class is a specialisation of the Engine class. It implements a file-based server to be used for testing
    """

    def __init__(self, config: EngineConfig, auth: Auth):
        super(FileBasedEngine, self).__init__(config, auth)
        self._listening_list = []
        logger.warning("TEST MODE")
        self._polling_interval = 1  # for testing we can do a much faster polling time
        self._host = "localhost"
        self._port = ""

    def pull(
        self,
        key: str,
        key_only: bool = False,
        rev: int = None,
        prefix: bool = True,
        min_rev: int = None,
        max_rev: int = None,
    ) -> List[Dict[str, any]]:
        """
        This method implements a query to the notification server for all the key-values associated to the key as input.
        This key by default is a prefix, it can therefore return a set of key-values
        :param key: input in the query
        :param key_only: ignored for TestEngine
        :param rev: ignored for TestEngine
        :param prefix: if true the function will retrieve all the KV pairs starting with the key passed
        :param min_rev: ignored for TestEngine
        :param max_rev: ignored for TestEngine
        :return: List of key-value pairs formatted as dictionary
        """
        if key_only:
            logger.warning("key_only option is disabled in TestMode")
        if rev:
            logger.warning("rev option is disabled in TestMode")
        if min_rev:
            logger.warning("min_rev option is disabled in TestMode")
        if max_rev:
            logger.warning("max_rev option is disabled in TestMode")

        def read_key(k):
            try:
                with open(k, "r") as f:
                    v = f.read()
            except Exception:
                logger.warning(f"Reading of the {k} has failed")
                logger.debug("", exc_info=True)
                return
            new_kv = {"key": k, "value": v.encode()}
            new_kvs.append(new_kv)
            logger.debug(f"Key: {k} pulled successfully")

        logger.debug(f"Calling pull for {key}...")
        new_kvs: List[Dict[str, bytes]] = []
        if os.path.exists(key):
            if os.path.isdir(key):
                # first list the directory
                for x in os.walk(key):
                    for fp in x[2]:  # any file
                        kk: str = os.path.join(x[0], fp)
                        read_key(kk)
                    if not prefix:  # we only look at the current directory, nothing deeper
                        break
            else:
                read_key(key)

        logger.debug(f"Query for {key} completed")
        logger.debug(f"{len(new_kvs)} keys found")
        return new_kvs

    def delete(self, key: str, prefix: bool = True) -> List[Dict[str, bytes]]:
        """
        This method deletes all the keys associated to this key, the key is a prefix as default
        :param key: key prefix to delete
        :param prefix: if true the function will delete all the KV pairs starting with the key passed
        :return: kvs deleted
        """
        logger.debug(f"Calling delete for {key}...")

        del_kvs: List[Dict[str, bytes]] = []
        if os.path.exists(key):
            if os.path.isdir(key):
                # first list the directory
                for x in os.walk(key):
                    for fp in x[2]:  # any file
                        k: str = os.path.join(x[0], fp)
                        new_kv = {"key": k}
                        del_kvs.append(new_kv)
            else:
                new_kv = {"key": key}
                del_kvs.append(new_kv)

            # now the delete the directory or file
            try:
                if os.path.isdir(key):
                    rmtree(key)
                else:
                    os.remove(key)
            except Exception as e:
                logger.warning(f"Cannot delete the key {key},  {e}")
                logger.debug("", exc_info=True)

        logger.debug(f"Delete request for key {key} completed")

        return del_kvs

    def push(self, kvs: List[Dict[str, any]], ks_delete: List[str] = None, ttl: int = None) -> bool:
        """
        Method to submit a list of key-value pairs and delete a list of keys from the server as a single transaction
        :param kvs: List of KV pair
        :param ks_delete: List of keys to delete before the push of the new ones. Note that each key is read as a folder
        :param ttl: Not supported in this implementation
        :return: True if successful
        """
        logger.debug("Calling push...")

        # first delete the keys requested
        if ks_delete is not None and len(ks_delete) != 0:
            for kd in ks_delete:
                if os.path.exists(kd):
                    try:
                        if os.path.isdir(kd):
                            rmtree(kd)
                        else:
                            os.remove(kd)
                    except Exception as e:
                        logger.warning(f"Cannot delete the key {kd},  {e}")
                        logger.debug("", exc_info=True)

        # save the keys to files
        for kv in kvs:
            k = kv["key"]
            v = kv["value"]
            file_name: str = k.split("/").pop()
            if not file_name == "":
                folder_path = k[: -len(file_name)]
            else:  # if k ends in / it means it the base directory, this is used to saved the status
                folder_path = k
                k += "status"
            if not os.path.exists(folder_path):
                try:
                    os.makedirs(folder_path, exist_ok=True)
                except OSError:
                    logger.warning(f"Cannot create the directory: {folder_path}")
                    logger.debug("", exc_info=True)
                    return False
            try:
                with open(k, "w+") as f:
                    f.write(v)
            except Exception:
                logger.warning(f"Saving of the {k} has failed")
                logger.debug("", exc_info=True)
                return False

        logger.debug("Transaction completed")

        return True

    def _polling(
        self,
        key: str,
        callback: callable([str, str]),
        channel: Queue,
        from_date: datetime = None,
        to_date: datetime = None,
    ):
        """
        This method implements the active polling using watchdog
        :param key: key to watch as a prefix
        :param callback: function to call if any change happen
        :param channel: global communication channel among threads
        :param from_date: ignored for TestMode
        :param to_date: ignored for TestMode
        :return:
        """
        if from_date:
            logger.warning("from_date option is disabled in TestMode")
        if to_date:
            logger.warning("to_date option is disabled in TestMode")

        try:
            # Create the directory to watch if it doesn't exist
            if not os.path.exists(key):
                try:
                    os.makedirs(key, exist_ok=True)
                except OSError:
                    logger.warning(f"Cannot create the directory: {key}")
                    logger.debug("", exc_info=True)
                    return False

            # Define a class to handle the new events using watchdog
            class EventHandler(FileSystemEventHandler):
                def __init__(self, engine, debounce_time=0.5, polling_interval=1):
                    super().__init__()
                    self._engine = engine
                    self._last_processed_time = 0
                    self._debounce_time = debounce_time
                    self._polling_interval = polling_interval

                def on_modified(self, event):
                    if not event.is_directory:
                        kvs = self._engine.pull(key=event.src_path)
                        for kv in kvs:
                            k = kv["key"]
                            v = kv["value"].decode()
                            if kv["key"].endswith("status"):
                                continue
                            logger.debug(f"Notification received for key {k}")
                            try:
                                callback(k, v)
                            except Exception as ee:
                                logger.error(f"Error with notification trigger, exception: {type(ee)} {ee}")
                                logger.debug("", exc_info=True)
                    time.sleep(self._polling_interval)

            # Set up the observer and event handler
            event_handler = EventHandler(engine=self)
            observer = Observer()
            observer.schedule(event_handler, key, recursive=True)
            observer.start()

            # Daemon thread to allow stopping
            def t_run():
                observer.join()

            t = threading.Thread(target=t_run)
            t.setDaemon(True)
            t.start()

            # Stop condition
            while key in self._listeners:
                time.sleep(0.1)
            observer.stop()
            observer.join()

        except Exception as e:
            logger.error(f"Error while listening to key {key}, {e}")
            logger.debug("", exc_info=True)
            _thread.interrupt_main()



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from .type_handler import TypeHandler


class StringHandler(TypeHandler):
    def __init__(self, key, required=False, canonic=None):
        super(StringHandler, self).__init__(key, required)
        self._canonic = canonic

    @property
    def canonic(self) -> str:
        return self._canonic

    def valid(self, value: any) -> bool:
        return True

    def canonise(self, value: any) -> str:
        value_str = str(value)
        if self.canonic is not None:
            if self.canonic == "lower":
                return value_str.lower()
            elif self.canonic == "upper":
                return value_str.upper()
            else:
                raise AttributeError(f"Case {self.canonic} not recognised")
        else:
            return value_str



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from typing import Dict

from .. import logger
from .trigger import Trigger, TriggerType


class TriggerFactory:
    """
    This class is in charge of creating a trigger for the specific trigger requested.
    """

    def create_trigger(self, notification: Dict[str, any], params: Dict[str, any]) -> Trigger:
        assert "type" in params, "'type' is a mandatory field in trigger"
        # find specific trigger class
        trigger_type = TriggerType[params.get("type").lower()]
        trigger_class = trigger_type.get_class()

        # instantiate the specific trigger
        logger.debug(f"Creating {trigger_type.name} trigger...")
        t = trigger_class(notification, params)
        logger.debug(f"Trigger {trigger_type.name} created")

        return t



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import json
import os
from enum import Enum
from typing import Dict

from .. import HOME_FOLDER, SYSTEM_FOLDER, logger
from ..custom_exceptions import ServiceConfigException

LOCAL_SCHEMA_FOLDER = "service_configuration"
LISTENER_SCHEMA_FILE_NAME = "event_listener_schema.json"
DEFAULT_SCHEMA_FILE_NAME = "default_listener_schema.json"


class ListenerSchemaParserType(Enum):
    """
    This Enum describes the various notification mechanism currently available. This identifies the type of server
    this client is connecting to.
    """

    GENERIC = "ListenerSchemaParser"
    ECMWF = "EcmwfSchemaParser"

    def parser(self):
        return eval(self.value + "()")


class ListenerSchemaParser:
    def load(self, config):
        """
        This method implements the workflow regarding loading the listener schema file
        from a remote location or local location or default.

        Args:
            config (UserConfig): main configuration

        Returns:
            Dict: Event Listener schema
        """
        logger.debug("Updating schema configurations...")
        local_schema_file_paths = []
        remote_schema_files = []
        if config.remote_schema:
            # Pull the latest event listeners configuration files
            from ..service_config_manager import ServiceConfigManager

            config_manager = ServiceConfigManager(config)
            remote_schema_files = config_manager.pull(config.notification_engine.service)
        else:
            # First the system config file
            system_path = os.path.join(SYSTEM_FOLDER, LOCAL_SCHEMA_FOLDER)
            # Check the directory exist
            if not os.path.exists(system_path):
                local_schema_file_paths = self._scan_folder(system_path)
            else:
                logger.debug(f"Schema folder in {system_path} not found")

            # Second the Home config file
            home_path = os.path.join(os.path.expanduser(HOME_FOLDER), LOCAL_SCHEMA_FOLDER)
            # Check the directory exist
            if os.path.exists(home_path):
                local_schema_file_paths = self._scan_folder(home_path)
            else:
                logger.debug(f"Schema folder in {home_path} not found")

        # parse the file loaded
        return self.parse(local_schema_file_paths, remote_schema_files)

    def _scan_folder(self, directory):
        files = []
        for x in os.walk(directory):
            for fp in x[2]:
                files.append(os.path.join(directory, fp))
        return files

    def _load_default_schema(self):
        default_path = os.path.join(os.path.dirname(__file__), DEFAULT_SCHEMA_FILE_NAME)
        with open(default_path) as evl_json:
            evl_schema = json.load(evl_json)
        return evl_schema

    def _load_event_listener_schema(self, local_schema_file_paths, remote_schema_files):
        # first check if we have remote schema. This takes priority
        if len(remote_schema_files) == 0:
            logger.debug("No remote schema file found")

            # then check if we have local schema
            if len(local_schema_file_paths) == 0:
                logger.warning("Not local listener schema file found, using default schema")
                # load the default schema
                evl_schema = self._load_default_schema()
            else:
                # search for the schema in the local folder
                evl_schema_file_path = list(
                    filter(lambda esfp: LISTENER_SCHEMA_FILE_NAME in esfp, local_schema_file_paths)
                )
                if len(evl_schema_file_path) != 1:
                    raise ServiceConfigException("No local event listener schema file found")
                with open(evl_schema_file_path[0]) as evl_json:
                    evl_schema = json.load(evl_json)
        else:
            # search for the schema among the remote schema
            evl_schema_file_s = list(filter(lambda esf: LISTENER_SCHEMA_FILE_NAME in esf["key"], remote_schema_files))
            if len(evl_schema_file_s) != 1:
                raise ServiceConfigException("No remote event listener schema file found")
            evl_schema_file = evl_schema_file_s[0]
            evl_schema = json.loads(evl_schema_file["value"].decode())

        return evl_schema

    def parse(self, local_schema_file_paths, remote_schema_files):
        """
        This methods takes all schema files found and parse them to create the final schema file.
        Custom parser should re-implement this method.

        Args:
            local_schema_file_paths (List): List of file paths to the
            remote_schema_files (List): [description]

        Returns:
            Dict: Event Listener schema
        """
        evl_schema = self._load_event_listener_schema(local_schema_file_paths, remote_schema_files)

        return evl_schema


class EcmwfSchemaParser(ListenerSchemaParser):
    MARS_SCHEMA_FILE_NAME = "language.json"

    def parse(self, local_schema_file_paths, remote_schema_files):
        """
        This re-implements the parent parse method to include the enum values from the MARS language schema
        """

        # load event listener schema
        evl_schema = self._load_event_listener_schema(local_schema_file_paths, remote_schema_files)

        # load mars schema
        mars_schema = self._load_mars_schema(remote_schema_files)

        # merge the schemas
        listener_schema = self._update_schema(mars_schema, evl_schema)
        return listener_schema

    def _load_mars_schema(self, schema_files):
        mars_schema_file_s = list(filter(lambda mf: self.MARS_SCHEMA_FILE_NAME in mf["key"], schema_files))
        if len(mars_schema_file_s) != 1:
            raise ServiceConfigException("No MARS language schema file found")
        mars_schema_file = mars_schema_file_s[0]
        mars_schema = json.loads(mars_schema_file["value"].decode())

        return mars_schema

    def _update_schema(self, mars_schema, evl_schema) -> Dict[str, any]:
        """
        This method is used to update the aviso schema with values from the mars schema and the ones from the custom
        schema defined in the config file.
        :param mars_schema - this is the MARS language schema
        :param evl_schema - this is the Event Listener schema to update
        :return: aviso schema updated
        """
        # search for all the enum keys in all event listener types and add the mars values to it
        logger.debug("Parsing mars language schema...")
        for e in evl_schema.items():
            if isinstance(e[1], dict) and e[1].get("request"):
                request = e[1].get("request")
                for k in request:
                    for t in request[k]:
                        if t["type"] == "EnumHandler":
                            if "values" not in t:
                                t["values"] = []
                            # check the mars schema
                            mars_enums = mars_schema["_field"][k]["values"]
                            for me in mars_enums:
                                if isinstance(me, list):
                                    for en in me:
                                        t["values"].append(en)
                                else:
                                    t["values"].append(me)
                logger.debug("Parsing completed")

        return evl_schema



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from .type_handler import TypeHandler


class FloatHandler(TypeHandler):
    def __init__(self, key, required=False, canonic=None):
        super(FloatHandler, self).__init__(key, required)
        self._canonic = canonic

    @property
    def canonic(self) -> str:
        return self._canonic

    def valid(self, value: any) -> bool:
        try:
            value = float(value)
        except ValueError as e:
            raise ValueError(f"Key {self.key} has to be a float", e)
        return True

    def canonise(self, value: any) -> str:
        value = float(value)
        if self.canonic is not None:
            return self.canonic.format(value)
        else:
            return str(value)



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import io
import pathlib
import re

from setuptools import find_packages, setup

__version__ = re.search(
    r'__version__\s*=\s*[\'"]([^\'"]*)[\'"]',  # It excludes inline comment too
    io.open("pyaviso/version.py", encoding="utf_8_sig").read(),
).group(1)

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

INSTALL_REQUIRES = (pathlib.Path(__file__).parent / "requirements.txt").read_text().splitlines()

setup(
    name="pyaviso",
    author="ECMWF",
    author_email="software.support@ecmwf.int",
    license="Apache-2.0",
    url="https://github.com/ecmwf/aviso",
    description="Time-critical notification system designed to trigger users' workflows across HPC and Cloud systems",
    long_description=long_description,
    long_description_content_type="text/markdown",
    version=__version__,
    packages=find_packages(exclude=("tests", "aviso-server")),
    include_package_data=True,
    install_requires=INSTALL_REQUIRES,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Natural Language :: English",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    entry_points={"console_scripts": ["aviso=pyaviso.cli_aviso:cli", "aviso-config=pyaviso.cli_aviso_config:cli"]},
)



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from abc import ABC, abstractmethod

from ..user_config import UserConfig


class Auth(ABC):
    """
    Abstraction for the authentication mechanism used to talk with the aviso server.
    """

    @staticmethod
    def get_auth(config: UserConfig):
        """Static method return a instance of the authenticator"""
        return config.auth_type.get_class()(config)

    def __init__(self):
        super(Auth, self).__init__()

    @property
    def username(self):
        return None

    @property
    def password(self):
        return None

    @abstractmethod
    def header(self):
        """
        :return: the header to add to the REST calls to comply with the authentication protocol
        """
        pass



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import json
from typing import Dict

from .. import logger
from . import trigger
from .trigger import TriggerType


class EchoTrigger(trigger.Trigger):
    """
    This class implements the 'Echo' trigger by printing out to the system log the
    notification
    """

    def __init__(self, notification: Dict[str, any], params: Dict[str, any]):
        trigger.Trigger.__init__(self, notification, params)
        self.trigger_type = TriggerType.echo

    def execute(self):
        logger.info("Starting Echo Trigger...")
        logger.info("Notification received:")
        print(json.dumps(self.notification, indent=4, sort_keys=True))
        logger.debug(f"{self.notification}")
        logger.info("Echo Trigger completed")



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import getpass
import json
import os
import threading
from abc import ABC, abstractmethod
from datetime import datetime
from queue import Queue
from typing import Dict, List

from .. import __version__, exit_channel, logger
from ..authentication.auth import Auth
from ..user_config import EngineConfig
from . import EngineType

DATE_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"


class Engine(ABC):
    """
    This class implements the interface to the notification server. It is abstract to separate the interface from the
    specific notification server
    """

    def __init__(self, config: EngineConfig, auth: Auth):
        super(Engine, self).__init__()
        self._host = config.host
        self._port = config.port
        self._polling_interval = config.polling_interval
        self._engine_type = config.type
        self.timeout = config.timeout
        self.catchup = config.catchup
        self._auth = auth
        self._https = config.https
        self.automatic_retry_delay = config.automatic_retry_delay
        self._listeners = []
        # this is used to synchronise multiple listening threads accessing the state
        self._state_lock = threading.Lock()
        # this is used to synchronise multiple listening threads accessing the listeners list
        self._listeners_lock = threading.Lock()

    @property
    def engine_type(self) -> EngineType:
        return self._engine_type

    @property
    def host(self) -> str:
        return self._host

    @property
    def port(self) -> int:
        return self._port

    @property
    def auth(self) -> Auth:
        return self._auth

    @property
    def https(self) -> bool:
        return self._https

    @abstractmethod
    def pull(
        self,
        key: str,
        key_only: bool = False,
        rev: int = None,
        prefix: bool = True,
        min_rev: int = None,
        max_rev: int = None,
    ) -> List[Dict[str, any]]:
        """
        Abstract method to query the notification server for all the key-values associated to the key as input.
        This key can be a prefix, it can therefore return a set of key-values
        :param key: input in the query
        :param key_only: if True no values are returned
        :param rev: revision to pull
        :param prefix: if true the function will retrieve all the KV pairs starting with the key passed
        :param min_rev: if provided it filters for only KV pairs with mod_revision >= to min_rev
        :param max_rev: if provided it filters for only KV pairs with mod_revision <= to max_rev
        :return: List of key-value pairs formatted as dictionary
        """
        pass

    @abstractmethod
    def push(self, kvs: List[Dict[str, any]], ks_delete: List[str] = None, ttl: int = None) -> bool:
        """
        Abstract method to submit a list of key-value pairs and delete a list of keys from the server as a
        single transaction
        :param kvs: List of KV pair
        :param ks_delete: List of keys to delete before the push of the new ones. Note that each key is read as a folder
        :param ttl: time to leave of the keys pushed, once expired the keys will be deleted
        :return: True if successful
        """
        pass

    @abstractmethod
    def delete(self, key) -> List[Dict[str, bytes]]:
        """
        This method deletes all the keys associated to this prefix
        :param key: key prefix to delete
        :return: kvs deleted
        """
        pass

    @abstractmethod
    def _polling(
        self,
        key: str,
        callback: callable([str, str]),
        channel: Queue,
        from_date: datetime = None,
        to_date: datetime = None,
    ):
        """
        This method implements the active polling
        :param key: key to watch as a prefix
        :param callback: function to call if any change happen
        :param channel: global communication channel among threads
        :param from_date: date from when to request notifications, if None it will be from now
        :param to_date: date until when to request notifications, if None it will be until now
        :return:
        """
        pass

    def listen(
        self, keys: List[str], callback: callable([str, str]), from_date: datetime = None, to_date: datetime = None
    ) -> bool:
        """
        This method allows to listen for changes to specific keys. Note that the key is always considered as a prefix.
        The listening is implemented with a background thread doing a active polling. Multiple listening threads can
        be created by calling this method multiple times.

        :param keys: keys to watch
        :param callback: function to trigger in case of changes
        :param from_date: date from when to request notifications, if None it will be from now
        :param to_date: date until when to request notifications, if None it will be until now
        :return: True if the listener is in execution, False otherwise
        """
        logger.debug("Calling listen...")
        for key in keys:
            try:
                # create a background thread for the polling
                t = threading.Thread(target=self._polling, args=(key, callback, exit_channel, from_date, to_date))
                t.setDaemon(True)
                # adding the thread to the global list
                logger.debug(f"Starting thread to listen to {key}")
                self._add_listener(key)
                t.start()
                logger.debug(f"Thread {t.ident} started to listen to {key}")
            except Exception as e:
                logger.error(f"Error in listening to {key}: {e}")
                logger.debug("", exc_info=True)
                self._remove_listener(key)
                return False
        return True

    def stop(self, key: str = None) -> bool:
        """
        This method is used to stop a listening thread. if no key is provided all the listening thread will be stopped

        :param key: the key associated to the listening thread
        :return: True if the listener is cancelled, False otherwise
        """
        logger.debug("Calling stop...")
        if len(self._listeners) > 0:
            if key is None:  # if not key is defined we simply stop all the listeners
                # by removing its entry from this list the thread will automatically stop
                logger.debug("Stopping all the polling")
                self._remove_all_listeners()
                return True
            elif key in self._listeners:
                # stop the polling
                logger.debug(f"Stopping the polling for key {key}")
                # by removing its entry from this list the thread will automatically stop
                self._remove_listener(key)
                return True
            else:
                logger.debug(f"Cannot find polling of key {key}")
                return False
        return True

    def push_with_status(
        self,
        kvs: List[Dict[str, any]],
        base_key: str,
        message: str = "",
        admin_key: str = None,
        ks_delete: List[str] = None,
        ttl: int = None,
    ) -> bool:
        """
        Method to submit a list of key-value pairs and delete a list of keys from the server as a
        single transaction. This method also updates the status of the base key.
        :param kvs: List of KV pair
        :param base_key: base key where to push the status
        :param message: message to be part of the status update
        :param admin_key: admin key to push together with the status
        :param ks_delete: List of keys to delete before the push of the new ones. Note that each key is read as a folder
        :param ttl: time to leave of the keys pushed, once expired the keys will be deleted
        :return: True if successful
        """
        # create the status payload
        status = {
            "etcd_user": getattr(self.auth, "username", None),
            "message": message,
            "unix_user": getpass.getuser(),
            "aviso_version": __version__,
            "engine": self._engine_type.name,
            "hostname": os.uname().nodename,
            "date_time": datetime.utcnow().strftime(DATE_FORMAT),
        }

        # update the status with the revision of the current status. This helps creating a linked list
        old_status_kvs = self.pull(base_key, prefix=False)
        if len(old_status_kvs) == 1:
            self._status_as_linked_list(status, old_status_kvs)

        status_kv = {"key": base_key, "value": json.dumps(status)}  # push it as a json
        kvs.append(status_kv)

        if admin_key:
            # prepare the admin key value pair
            admin_kv = {"key": admin_key, "value": "None"}
            kvs.append(admin_kv)

        return self.push(kvs, ks_delete, ttl)

    def _status_as_linked_list(self, new_status, old_status_kvs):
        if "mod_rev" in old_status_kvs[0]:  # test engine does not have it
            new_status["prev_rev"] = old_status_kvs[0]["mod_rev"]

            # update the status with date and rev of the last_prev_day. This helps creating a linked list across days
            old_status = json.loads(old_status_kvs[0]["value"].decode())
            old_date_time = datetime.strptime(old_status["date_time"], DATE_FORMAT)
            new_date_time = datetime.strptime(new_status["date_time"], DATE_FORMAT)

            # compare status dates
            if old_date_time.date() == new_date_time.date():
                # we are still on the same day
                if "last_prev_day_rev" in old_status:
                    new_status["last_prev_day_rev"] = old_status["last_prev_day_rev"]
            else:
                # new day -> use the previous status as last_prev_day
                new_status["last_prev_day_rev"] = new_status["prev_rev"]

    def _add_listener(self, key: str):
        with self._listeners_lock:
            self._listeners.append(key)

    def _remove_all_listeners(self):
        with self._listeners_lock:
            self._listeners.clear()

    def _remove_listener(self, key: str):
        with self._listeners_lock:
            self._listeners.remove(key)



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from datetime import datetime
from typing import Dict, List, Tuple

import yaml

from . import exit_channel, logger, user_config
from .authentication.auth import Auth
from .custom_exceptions import EventListenerException, InvalidInputError
from .engine import engine_factory as ef
from .event_listeners.event_listener import DEFAULT_PAYLOAD_KEY, EventListener
from .event_listeners.listener_manager import ListenerManager


class NotificationManager:
    """
    This class manages implements the various operations associated to the notification system
    """

    def __init__(self):
        self.listener_manager = ListenerManager()

    def _listen(
        self,
        config: user_config.UserConfig = None,
        listeners_file_paths: List[str] = None,
        listeners: Dict[str, any] = None,
        from_date: datetime = None,
        to_date: datetime = None,
    ) -> int:
        """
        This method parses the inputs and calls the listener manager to create the listeners
        :param config: UserConfig object
        :param listeners_file_paths: list of file paths to YAML listener files
        :param listeners: listeners as dictionaries
        :param from_date: date from when to request notifications, if None it will be from now
        :param to_date: date until when to request notifications, if None it will be until now
        :return: number of listeners running
        """
        # check we have listeners
        listeners_list = []
        if listeners_file_paths is not None and len(listeners_file_paths) > 0:
            listeners_list = self._load_listener_files(listeners_file_paths)
        elif listeners is None:
            if hasattr(config, "listeners"):
                listeners_list.append(config.listeners)
            else:
                raise EventListenerException("Listeners not defined")
        else:
            listeners_list.append(listeners)

        # retrieve listener schema
        listener_schema = config.schema_parser.parser().load(config)

        # Call the listener manager
        return self.listener_manager.listen(listeners_list, listener_schema, config, from_date, to_date)

    def listen(
        self,
        config: user_config.UserConfig = None,
        listeners_file_paths: List[str] = None,
        listeners: Dict[str, any] = None,
        from_date: datetime = None,
        to_date: datetime = None,
        now: bool = False,
        catchup: bool = False,
    ):
        """
        This method implements the main workflow to instantiate and execute new listeners and holding the main thread in
        active waiting.
        :param config: UserConfig object
        :param listeners_file_paths: list of file paths to YAML listener files
        :param listeners: listeners as dictionaries
        :param from_date: date from when to request notifications, if None it will be from now
        :param to_date: date until when to request notifications, if None it will be until now
        :param now: if True ignore missed notifications, only listen to new ones
        :param catchup: if True retrieve first the missed notifications
        :return:
        """
        logger.debug("Calling listen...")

        # first check the config
        if config is None:
            config = user_config.UserConfig()

        # check the inputs
        now_date = datetime.utcnow()
        if from_date:
            assert from_date < now_date, "from_date must be in the past"
        if to_date:
            assert to_date < now_date, "to_date must be in the past"
            assert from_date is not None, "from_date is required if to_date is defined"
            assert to_date > from_date, "to_date must be later than from_date"

        # define the catchup behaviour and set it in the notification engine
        assert not (now and catchup), "Only now or catchup can be specified at the same time"
        if catchup:
            config.notification_engine.catchup = True
        else:
            if now:
                config.notification_engine.catchup = False

        # Call the listener manager
        self._listen(config, listeners_file_paths, listeners, from_date, to_date)

        # keep the main process running and wait for the listening thread to terminate
        l_exit = exit_channel.get()  # this is blocking until all listener ends or there is an error
        if l_exit:  # it exits successful
            return
        else:  # it exits with errors
            raise EventListenerException("Error in one of the listening process")

    def key(
        self, params: Dict, config: user_config.UserConfig = None, listener_schema: Dict = None
    ) -> Tuple[str, str, str]:
        """
        Generate a key to send to the notification server with the params passed and complying to the current schema
        :param params: parameters to use in the key
        :param config:
        :param listener_schema: event listener schema are loaded as dictionary
        :return: tuple of leaf key and root key
        """
        logger.debug(f"Calling generate key with the following parameters {params}...")

        # first check the config
        if config is None:
            config = user_config.UserConfig()

        if "event" not in params:
            raise InvalidInputError("Invalid notification, 'event' could not be located")
        listener_type = params["event"]

        if not listener_schema:
            # retrieve listener schema
            logger.debug("Getting schema...")
            listener_schema = config.schema_parser.parser().load(config)
        # # extract the relevant listener schema
        if listener_type not in listener_schema:
            raise InvalidInputError(f"Invalid notification, {listener_type} could not be located in the schema")
        event_schema = listener_schema.get(listener_type)
        logger.debug("Relevant schema successfully found")

        logger.debug("Generating key...")
        filtered_params = params.copy()
        filtered_params.pop("event")
        key, root, admin_key = EventListener.derive_notification_keys(
            filtered_params, event_schema, config.notification_engine.type
        )
        logger.debug(f"Keys generated {root}, {key}, {admin_key}")

        return key, root, admin_key

    def value(self, params: Dict, config: user_config.UserConfig = None) -> str:
        """
        :param params: parameters to use in the key
        :param config:
        :return: the value on the server corresponding to the key which is generated according to the current schema and
        the parameters defined
        """
        logger.debug(f"Calling get value with the following parameters {params}...")

        # first check the config
        if config is None:
            config = user_config.UserConfig()

        # first generate the key corresponding the to the parameters passed
        key, base_key, admin_key = self.key(params, config)

        # create the engine
        engine_factory: ef.EngineFactory = ef.EngineFactory(config.notification_engine, Auth.get_auth(config))
        engine = engine_factory.create_engine()

        # retrieve the value
        kvs = engine.pull(key=key, prefix=False)
        assert len(kvs) < 2, "Error in retrieving value from key, more than one value returned"
        if len(kvs) == 0:
            logger.debug("No value returned")
        else:
            return kvs[0]["value"].decode()

    def notify(self, notification: Dict, config: user_config.UserConfig = None) -> bool:
        """
        Send a notification to the server. The notification is made of a key-value pair created using the params passed
        and a status that is sent to the base key. This is needed for the catchup feature.
        :param notification: dictionary of the notification ready to submit
        :param config: UserConfig object
        :return: True if the notification has been submitted
        """
        logger.debug(f"Calling notify with the following notification {notification}...")

        # first check the config
        if config is None:
            config = user_config.UserConfig()

        # retrieve listener schema
        logger.debug("Getting schema...")
        listener_schema = config.schema_parser.parser().load(config)

        # validate the input
        try:
            # check the payload key
            payload_key = listener_schema.get("payload")
            if not payload_key:
                payload_key = DEFAULT_PAYLOAD_KEY
            assert "event" in notification, "Invalid notification, 'event' could not be located"
            value = "None"
            if payload_key in notification:
                value = notification.pop(payload_key)
        except AssertionError as e:
            raise InvalidInputError(e)

        # first check the config
        if config is None:
            config = user_config.UserConfig()

        # read the TTL for this key
        ttl = config.key_ttl
        if "ttl" in notification:
            ttl = int(notification.pop("ttl"))
        # translate default -1 into the default of the engine interface
        ttl = None if ttl == -1 else ttl

        # generate the key
        key, base_key, admin_key = self.key(notification, config, listener_schema)

        # create the engine
        engine_factory: ef.EngineFactory = ef.EngineFactory(config.notification_engine, Auth.get_auth(config))
        engine = engine_factory.create_engine()

        # submit the notification with status update
        logger.debug(f"Submit key {key}, value {value} with status update")
        kvs = [{"key": key, "value": value}]
        engine.push_with_status(
            kvs, base_key=base_key, admin_key=admin_key, message=f"notification to key {key}", ttl=ttl
        )

        return True

    def _load_listener_files(self, listener_files: List[str]):
        """
        :param listener_files: list of file paths to YAML listener files
        :return: return list of dictionaries from the listener files
        """
        # Open YAML notification listener files
        assert listener_files is not None
        listeners: List[Dict[str, any]] = []
        for listener_file in listener_files:
            logger.debug(f"Reading listener file {listener_file}")
            try:
                with open(listener_file, "r") as f:
                    listeners_dict = yaml.safe_load(f)
                    listeners.append(listeners_dict)
            except Exception as e:
                raise EventListenerException(f"Not able to load listener file {listener_file},{e}")
        return listeners



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import itertools
import re
from datetime import datetime
from typing import Dict, List

import parse

from .. import logger
from ..custom_exceptions import EventListenerException
from ..engine import EngineType
from ..engine.engine import Engine
from ..triggers import trigger_factory as tf
from .validation import *  # noqa: F403

DEFAULT_PAYLOAD_KEY = "payload"


class EventListener:
    """
    This class contains of all the details needed to create an event listener and execute its triggers
    """

    def __init__(
        self,
        event_type: str,
        engine: Engine,
        request: Dict[str, any],
        triggers: List[Dict[str, any]],
        listener_schema: Dict[str, any],
        from_date: datetime = None,
        to_date: datetime = None,
        payload_key: str = None,
    ):
        self._event_type = event_type
        self._engine = engine
        self._request = request if request else {}
        self._triggers = triggers
        self._listener_schema = listener_schema
        self._trigger_factory = tf.TriggerFactory()
        self._keys = self.key_expansion(self._request)
        self._filter = self.filter_expansion(self._request)
        self._from_date = from_date
        self._to_date = to_date
        self.payload_key = payload_key

    def __str__(self):
        return f"{self.event_type} listener to keys: {self.keys}"

    @property
    def event_type(self) -> str:
        return self._event_type

    @property
    def payload_key(self) -> str:
        return self._payload_key

    @payload_key.setter
    def payload_key(self, payload_key: any):
        if payload_key:
            self._payload_key = payload_key
        else:
            self._payload_key = DEFAULT_PAYLOAD_KEY

    @property
    def from_date(self) -> datetime:
        return self._from_date

    @property
    def to_date(self) -> datetime:
        return self._to_date

    @property
    def engine(self) -> Engine:
        return self._engine

    @property
    def request(self) -> Dict[str, any]:
        return self._request

    @property
    def keys(self) -> List[str]:
        return self._keys

    @property
    def triggers(self) -> List[Dict[str, any]]:
        return self._triggers

    @property
    def listener_schema(self) -> Dict[str, any]:
        return self._listener_schema

    @property
    def trigger_factory(self) -> tf.TriggerFactory:
        return self._trigger_factory

    def key_expansion(self, request: Dict[str, any]) -> List[str]:
        """
        This functions composes the keys to watch using the listener request dictionary
        :param request:
        :return: List of keys
        """
        # read the key format from the schema
        key_base = EventListener._key_base_format(self.listener_schema, self.engine.engine_type)
        try:
            key: str = key_base.format(**request)
        except KeyError as e:
            raise KeyError(f"Wrong listener file: {','.join(e.args)} required")

        # compose multiple keys from the list values
        keys: List[str] = []
        template = r"\[[^\[\]]*\]"
        matches = re.findall(template, key)
        if len(matches) == 0:
            keys.append(key)
        else:
            # Build a list of the possible substitutions
            ldests = []
            for m in matches:
                ldests.append(eval(m))
            # Generate the various pairings
            for lproduct in itertools.product(*ldests):
                k = key
                for src, dest in itertools.zip_longest(matches, lproduct):
                    # Replace each term (you could optimise this using a single re.sub)
                    k = k.replace(src, dest)
                keys.append(k)
        return keys

    def filter_expansion(self, request: Dict[str, any]) -> Dict[str, List[any]]:
        """
        This method builds from the request the filter used in the post-notification phase
        :param request:
        :return: filter as a dictionary
        """
        # initialise the filter
        request_filter: Dict[str, List[any]] = {}
        # get the request schema
        assert "request" in self.listener_schema, "Wrong schema structure, 'request' could not be located"
        request_schema = self.listener_schema["request"]

        # validate and canonize the request
        EventListener._validate(request, request_schema)

        # scan through the request to create the filter object
        for r in request.keys():
            request_filter[r] = []
            value = request[r]
            if type(value) is list:
                for v in value:
                    request_filter[r].append(v)
            else:
                request_filter[r].append(value)

        return request_filter

    def parse_key(self, key: str) -> Dict[str, any]:
        """
        This function parses the key string received as part of the notification and create a dictionary out of it
        :param key:
        :return:
        """
        # read the key format from the schema
        key_put_format = EventListener._key_base_format(
            self.listener_schema, self.engine.engine_type
        ) + EventListener._key_stem_format(self.listener_schema, self.engine.engine_type)

        try:
            notification: Dict[str, any] = parse.parse(key_put_format, key, extra_types=[str]).named
        except AttributeError as e:
            logger.debug("", exc_info=True)
            raise EventListenerException(f"Key {key} failed validation, exception: {e}")
        return notification

    def callback(self, key: str, value: str):
        """
        This callback function first parses the key and build a notification dictionary, it then filters it using the
        self.filter requested. If it passes the filter phase the notification is then passed to the triggers
        otherwise the notification is ignored.
        :param key:
        :param value:
        :return:
        """
        # parse and filter the key
        not_request: Dict[str, any] = self.parse_key(key)

        if self._is_expected(not_request):
            # prepare the notification dictionary to pass to the trigger
            notification: Dict[str, any] = {"event": self.event_type, "request": not_request}
            if value != "None":
                notification[self.payload_key] = value
            # execute all the triggers defined in the EventListener
            logger.info("A valid notification has been received, executing triggers...")
            logger.debug(f"{notification}")
            self.execute_triggers(notification)

    def listen(self) -> bool:
        """
        This method is used to turn a EventListener object to an active notification request to the underlying
        notification mechanism. This method relies on the specific implementation of Engine class to activate the
        listener.

        :return: True if the listener is in execution, False otherwise
        """
        return self._engine.listen(self.keys, self.callback, self.from_date, self.to_date)

    def stop(self) -> bool:
        """
        This method is used to stop an active notification listener running on the underlying notification mechanism.

        :return: True if the listener has been cancelled
        """
        if self._engine.stop():
            logger.debug(f"{self} has been stopped")
            return True
        else:
            logger.warning(f"{self} not currently in execution")
            return False

    def execute_triggers(self, notification: Dict[str, any]):
        """
        This function is used to execute the triggers associated with this EventListener.
        :param notification:
        :return:
        """
        # execute all the triggers defined in the EventListener in order
        for t in self.triggers:
            try:
                # create the trigger
                trigger = self.trigger_factory.create_trigger(notification, t)
            except Exception as e:
                logger.error(f"Trigger {t} could not be created, {type(e)}: {e}")
                logger.debug("", exc_info=True)
                break  # the whole triggers execution stop
            else:  # run the trigger
                try:
                    trigger.execute()
                except Exception as e:
                    logger.error(f"Trigger {t} could not be executed,  {e}")
                    logger.debug("", exc_info=True)
                    break  # the whole triggers execution stop

    @staticmethod
    def derive_notification_keys(params: Dict[str, any], schema: Dict[str, any], engine_type: EngineType):
        """
        This function compose all the keys needed for a notification to the server using the parameters passed and
        the schema
        :param params:
        :param schema:
        :param engine_type
        :return: stem_key, base_key, admin_key
        """
        # get the request schema
        assert "request" in schema, "Wrong schema structure, 'request' could not be located"
        request_schema = schema["request"]
        # validate and canonize parameters
        EventListener._validate(params, request_schema)

        # read the base key format from the schema
        key_base_format = EventListener._key_base_format(schema, engine_type)
        try:
            # create the base key
            base_key: str = key_base_format.format(**params)
        except KeyError as e:
            raise KeyError(f"Wrong parameters: {','.join(e.args)} required")

        # read the stem key format from the schema
        key_stem_format = key_base_format + EventListener._key_stem_format(schema, engine_type)
        try:
            # create the stem key
            stem_key: str = key_stem_format.format(**params)
        except KeyError as e:
            raise KeyError(f"Wrong parameters: {','.join(e.args)} required")

        # read the admin key format from the schema
        key_admin_format = EventListener._key_admin_format(schema, engine_type)
        if key_admin_format:
            try:
                # create the admin key
                admin_key = key_admin_format.format(**params)
            except KeyError as e:
                raise KeyError(f"Wrong parameters: {','.join(e.args)} required")
        else:
            admin_key = None

        return stem_key, base_key, admin_key

    def _is_expected(self, notification: Dict) -> bool:
        """
        Helper method used to validate the notification received against the filters defined
        :param notification:
        :return:
        """
        expected = True
        for f_key, f_values in self._filter.items():
            assert f_key in notification, "Filter attribute not present in the notification"
            n_value = notification[f_key]
            # infer the type of the attributes from the filter type, the default is string
            if type(f_values[0]) is int:
                n_value = int(n_value)
            # now check if it matches the filters
            if n_value not in f_values:
                # notification does NOT complies with this filter attribute
                expected = False
                logger.debug(
                    f"Notification {notification} failed filter {f_key} with value {n_value} therefore it "
                    f"will be ignored"
                )
                break
        return expected

    @staticmethod
    def _key_base_format(schema, engine_type: EngineType) -> str:
        """
        Helper method used to extract the key base format from the schema
        :param schema:
        :param engine_type:
        :return: key base format
        """
        base_key_f = None
        assert "endpoint" in schema, "Wrong schema structure, 'endpoint' could not be located"
        endpoints = schema["endpoint"]
        assert len(endpoints) > 0, "Wrong schema structure, 'endpoint' should be a non empty list"
        for endpoint in endpoints:
            assert "engine" in endpoint, "Wrong schema structure, 'engine' in 'endpoint' could not be located"
            if engine_type.name.lower() in endpoint["engine"]:
                assert "base" in endpoint, "Wrong schema structure, 'base' in 'endpoint' could not be located"
                base_key_f = endpoint["base"]
                if not base_key_f.endswith("/"):
                    base_key_f = base_key_f + "/"
                break
        if base_key_f is None:
            raise EventListenerException("Key base could bot be located in the schema")
        return base_key_f

    @staticmethod
    def _key_stem_format(schema, engine_type: EngineType) -> str:
        """
        Helper method used to extract the key stem format from the schema
        :param schema:
        :param engine_type:
        :return: key stem format
        """
        stem_key_f = None
        assert "endpoint" in schema, "Wrong schema structure, 'endpoint' could not be located"
        endpoints = schema["endpoint"]
        assert len(endpoints) > 0, "Wrong schema structure, 'endpoint' should be a non empty list"
        for endpoint in endpoints:
            assert "engine" in endpoint, "Wrong schema structure, 'engine' in 'endpoint' could not be located"
            if engine_type.name.lower() in endpoint["engine"]:
                assert "stem" in endpoint, "Wrong schema structure, 'stem' in 'endpoint' could not be located"
                stem_key_f = endpoint["stem"]
                if stem_key_f.startswith("/"):
                    stem_key_f = stem_key_f[1 : len(stem_key_f)]
                break
        if stem_key_f is None:
            raise EventListenerException("Key base could bot be located in the schema")
        return stem_key_f

    @staticmethod
    def _key_admin_format(schema, engine_type: EngineType) -> str:
        """
        Helper method used to extract the key admin format from the schema
        :param schema:
        :param engine_type:
        :return: admin key format
        """
        admin_key_f = None
        assert "endpoint" in schema, "Wrong schema structure, 'endpoint' could not be located"
        endpoints = schema["endpoint"]
        assert len(endpoints) > 0, "Wrong schema structure, 'endpoint' should be a non empty list"
        for endpoint in endpoints:
            assert "engine" in endpoint, "Wrong schema structure, 'engine' in 'endpoint' could not be located"
            if engine_type.name.lower() in endpoint["engine"]:
                admin_key_f = endpoint.get("admin")
                break
        return admin_key_f

    @staticmethod
    def _validate(params, schema):
        """
        This private method validates and canonises the parameters passed using the schema.
        Note that the old params values are overwritten by the canonised values.
        :param params:
        :param schema:
        :return:
        """
        for p in params.keys():
            # check if this attribute is defined in the schema
            assert p in schema.keys(), f"Key {p} is not allowed"
            type_list = schema[p]
            valid = False
            for p_schema in type_list:
                try:
                    assert "type" in p_schema, f"Wrong schema structure, 'type' could not be located for {p}"
                    p_schema_c = p_schema.copy()
                    validator_class = p_schema_c.pop("type")
                    validator: TypeHandler = eval(f"{validator_class}(key=p, **p_schema_c)")  # noqa: F405
                    # format the values associated to this attribute
                    value = params[p]
                    if type(value) is list:
                        params[p] = []
                        for v in value:
                            params[p].append(validator.process(v))
                    else:
                        params[p] = validator.process(value)
                    # if no ValueError have been generated exit and don't valid against the other type handlers
                    valid = True
                    break
                except ValueError as e:
                    logger.debug(f"{e}")
            # check if at least one type handler was valid
            if not valid:
                raise ValueError(f"Value {params[p]} is not valid for key {p}")
